==> syd-3.21.0.tar.gz: syd-v3.21.0/src/syd-cp.rs <==
//
// syd: seccomp and landlock based application sandbox with support for namespaces
// src/syd-cp.rs: Utility to copy files using splice(2)
//
// Copyright (c) 2024 Ali Polatel <alip@chesswob.org>
//
// SPDX-License-Identifier: GPL-3.0-or-later

use core::sync::atomic::Ordering;
use std::{
    env,
    fs::File,
    os::fd::AsRawFd,
    path::Path,
    process::ExitCode,
    sync::atomic::AtomicBool,
    time::{Duration, Instant},
};

use anyhow::{Context, Result};
use getargs::{Opt, Options};
use nix::{
    errno::Errno,
    fcntl::{splice, SpliceFFlags},
    sys::{
        signal,
        signal::{SigAction, SigHandler, Signal},
    },
    unistd::isatty,
};

static INTERRUPTED: AtomicBool = AtomicBool::new(false);

extern "C" fn handle_sigint(_: nix::libc::c_int) {
    INTERRUPTED.store(true, Ordering::SeqCst);
}

fn process_data(src: &Path, dst: &Path, keeperm: bool, verbose: bool) -> Result<()> {
    let src_file = File::open(src).context("Error opening source file")?;
    let dst_path = if dst.is_dir() {
        let src_file_name = src.file_name().context("Error getting source file name")?;
        dst.join(src_file_name)
    } else {
        dst.to_path_buf()
    };

    let dst_file = File::create(&dst_path).context("Error creating destination file")?;

    if keeperm {
        // Get the file permissions of the source file
        let src_permissions = src_file
            .metadata()
            .context("Error getting metadata for source file")?
            .permissions();

        // Set the same permissions for the destination file
        dst_file
            .set_permissions(src_permissions.clone())
            .context("Error setting permissions for destination file")?;
    }

    let src_fd = src_file.as_raw_fd();
    let dst_fd = dst_file.as_raw_fd();

    let stime = Instant::now();
    let mut ltime = stime;
    let mut nbytes = 0;
    let mut nwrite = 0;
    let mut nprint = 0;
    let verbose = if verbose {
        isatty(nix::libc::STDERR_FILENO).unwrap_or(false)
    } else {
        false
    };

    while !INTERRUPTED.load(Ordering::Relaxed) {
        let n = match splice(
            src_fd,
            None,
            dst_fd,
            None,
            nix::libc::PIPE_BUF,
            SpliceFFlags::SPLICE_F_MORE,
        ) {
            Ok(0) | Err(Errno::EINTR) => break,
            Ok(n) => n,
            Err(errno) => return Err(errno.into()),
        };

        if !verbose {
            continue;
        }

        nbytes += n;
        nwrite += 1;

        let now = Instant::now();
        if now.duration_since(ltime) >= Duration::from_millis(500) {
            let elapsed = stime.elapsed();
            let speed = nbytes as f64 / elapsed.as_secs_f64();
            let output = format!(
                "{} bytes ({:.2} GB, {:.2} GiB) spliced, {:.2?} s, {:.2} MB/s",
                nbytes,
                nbytes as f64 / 1_000_000_000.0,
                nbytes as f64 / (1 << 30) as f64,
                elapsed,
                speed / (1 << 20) as f64
            );
            eprint!("{}{output}", "\x08".repeat(nprint));
            ltime = now;
            nprint = output.len();
        }
    }

    if verbose {
        let elapsed = stime.elapsed();
        eprintln!(
            "\n{} records of each {} bytes spliced.\n{} bytes ({:.2} GB, {:.2} GiB) spliced, {:.5?} s, {:.2} MB/s",
            nwrite,
            nix::libc::PIPE_BUF,
            nbytes,
            nbytes as f64 / 1_000_000_000.0,
            nbytes as f64 / (1 << 30) as f64,
            elapsed,
            nbytes as f64 / elapsed.as_secs_f64() / (1 << 20) as f64
        );
    }

    Ok(())
}

fn main() -> Result<ExitCode> {
    #[allow(clippy::disallowed_methods)]
    syd::set_sigpipe_dfl().expect("SIGPIPE");

    let args = env::args().skip(1).collect::<Vec<_>>();
    let mut opts = Options::new(args.iter().map(String::as_str));
    let mut keeperm = false;
    let mut verbose = false;

    #[allow(clippy::disallowed_methods)]
    while let Some(opt) = opts.next_opt().expect("next opt") {
        match opt {
            Opt::Short('h') => {
                help();
                return Ok(ExitCode::SUCCESS);
            }
            Opt::Short('a' | 'p') => {
                keeperm = true;
            }
            Opt::Short('v') => {
                verbose = true;
            }
            _ => {
                eprintln!("Unknown option: {opt:?}");
                help();
                return Ok(ExitCode::FAILURE);
            }
        }
    }

    let mut positionals = opts.positionals();
    let src = match positionals.next() {
        Some(src) => Path::new(src),
        None => {
            eprintln!("Error: Source file is required.");
            help();
            return Ok(ExitCode::FAILURE);
        }
    };

    let dst = match positionals.next() {
        Some(dst) => Path::new(dst),
        None => {
            eprintln!("Error: Destination file is required.");
            help();
            return Ok(ExitCode::FAILURE);
        }
    };

    // Set up the SIGINT handler
    unsafe {
        signal::sigaction(
            Signal::SIGINT,
            &SigAction::new(
                SigHandler::Handler(handle_sigint),
                signal::SaFlags::empty(),
                signal::SigSet::empty(),
            ),
        )
    }
    .context("Error setting up SIGINT handler")?;

    process_data(src, dst, keeperm, verbose)
        .map(|_| ExitCode::SUCCESS)
        .context("Error processing data")
}

fn help() {
    println!("Usage: syd-cp [-hapv] <src-file> <dst-file>");
    println!("Utility to copy files using splice(2)");
    println!("  -h        Display help.");
    println!("  -p, -a    Preserve file permissions.");
    println!("  -v        Enable verbose mode.");
}

==> syd-3.22.0.tar.gz: syd-v3.22.0/src/syd-cp.rs <==
//
// syd: seccomp and landlock based application sandbox with support for namespaces
// src/syd-cp.rs: Utility to copy files using splice(2)
//
// Copyright (c) 2024 Ali Polatel <alip@chesswob.org>
//
// SPDX-License-Identifier: GPL-3.0-or-later

use core::sync::atomic::Ordering;
use std::{
    env,
    fs::File,
    os::fd::AsRawFd,
    path::Path,
    process::ExitCode,
    sync::atomic::AtomicBool,
    time::{Duration, Instant},
};

use anyhow::{Context, Result};
use getargs::{Opt, Options};
use nix::{
    errno::Errno,
    fcntl::{splice, SpliceFFlags},
    sys::{
        signal,
        signal::{SigAction, SigHandler, Signal},
    },
    unistd::isatty,
};

static INTERRUPTED: AtomicBool = AtomicBool::new(false);

extern "C" fn handle_sigint(_: nix::libc::c_int) {
    INTERRUPTED.store(true, Ordering::SeqCst);
}

fn process_data(src: &Path, dst: &Path, keeperm: bool, verbose: bool) -> Result<()> {
    let src_file = File::open(src)
        .map_err(anyhow::Error::msg)
        .context("Error opening source file")?;
    let dst_path = if dst.is_dir() {
        let src_file_name = src
            .file_name()
            .ok_or_else(|| anyhow::Error::msg("Error getting source file name"))?;
        dst.join(src_file_name)
    } else {
        dst.to_path_buf()
    };

    let dst_file = File::create(&dst_path)
        .map_err(anyhow::Error::msg)
        .context("Error creating destination file")?;

    if keeperm {
        // Get the file permissions of the source file
        let src_permissions = src_file
            .metadata()
            .map_err(anyhow::Error::msg)
            .context("Error getting metadata for source file")?
            .permissions();

        // Set the same permissions for the destination file
        dst_file
            .set_permissions(src_permissions.clone())
            .map_err(anyhow::Error::msg)
            .context("Error setting permissions for destination file")?;
    }

    let src_fd = src_file.as_raw_fd();
    let dst_fd = dst_file.as_raw_fd();

    let stime = Instant::now();
    let mut ltime = stime;
    let mut nbytes = 0;
    let mut nwrite = 0;
    let mut nprint = 0;
    let verbose = if verbose {
        isatty(nix::libc::STDERR_FILENO).unwrap_or(false)
    } else {
        false
    };

    while !INTERRUPTED.load(Ordering::Relaxed) {
        let n = match splice(
            src_fd,
            None,
            dst_fd,
            None,
            nix::libc::PIPE_BUF,
            SpliceFFlags::SPLICE_F_MORE,
        ) {
            Ok(0) | Err(Errno::EINTR) => break,
            Ok(n) => n,
            Err(errno) => return Err(anyhow::Error::msg(errno)),
        };

        if !verbose {
            continue;
        }

        nbytes += n;
        nwrite += 1;

        let now = Instant::now();
        if now.duration_since(ltime) >= Duration::from_millis(500) {
            let elapsed = stime.elapsed();
            let speed = nbytes as f64 / elapsed.as_secs_f64();
            let output = format!(
                "{} bytes ({:.2} GB, {:.2} GiB) spliced, {:.2?} s, {:.2} MB/s",
                nbytes,
                nbytes as f64 / 1_000_000_000.0,
                nbytes as f64 / (1 << 30) as f64,
                elapsed,
                speed / (1 << 20) as f64
            );
            eprint!("{}{output}", "\x08".repeat(nprint));
            ltime = now;
            nprint = output.len();
        }
    }

    if verbose {
        let elapsed = stime.elapsed();
        eprintln!(
            "\n{} records of each {} bytes spliced.\n{} bytes ({:.2} GB, {:.2} GiB) spliced, {:.5?} s, {:.2} MB/s",
            nwrite,
            nix::libc::PIPE_BUF,
            nbytes,
            nbytes as f64 / 1_000_000_000.0,
            nbytes as f64 / (1 << 30) as f64,
            elapsed,
            nbytes as f64 / elapsed.as_secs_f64() / (1 << 20) as f64
        );
    }

    Ok(())
}

fn main() -> Result<ExitCode> {
    #[allow(clippy::disallowed_methods)]
    syd::set_sigpipe_dfl().expect("SIGPIPE");

    let args = env::args().skip(1).collect::<Vec<_>>();
    let mut opts = Options::new(args.iter().map(String::as_str));
    let mut keeperm = false;
    let mut verbose = false;

    #[allow(clippy::disallowed_methods)]
    while let Some(opt) = opts.next_opt().expect("next opt") {
        match opt {
            Opt::Short('h') => {
                help();
                return Ok(ExitCode::SUCCESS);
            }
            Opt::Short('a' | 'p') => {
                keeperm = true;
            }
            Opt::Short('v') => {
                verbose = true;
            }
            _ => {
                eprintln!("Unknown option: {opt:?}");
                help();
                return Ok(ExitCode::FAILURE);
            }
        }
    }

    let mut positionals = opts.positionals();
    let src = match positionals.next() {
        Some(src) => Path::new(src),
        None => {
            eprintln!("Error: Source file is required.");
            help();
            return Ok(ExitCode::FAILURE);
        }
    };

    let dst = match positionals.next() {
        Some(dst) => Path::new(dst),
        None => {
            eprintln!("Error: Destination file is required.");
            help();
            return Ok(ExitCode::FAILURE);
        }
    };

    // Set up the SIGINT handler
    unsafe {
        signal::sigaction(
            Signal::SIGINT,
            &SigAction::new(
                SigHandler::Handler(handle_sigint),
                signal::SaFlags::empty(),
                signal::SigSet::empty(),
            ),
        )
    }
    .map_err(anyhow::Error::msg)
    .context("Error setting up SIGINT handler")?;

    process_data(src, dst, keeperm, verbose)
        .map(|_| ExitCode::SUCCESS)
        .context("Error processing data")
}

fn help() {
    println!("Usage: syd-cp [-hapv] <src-file> <dst-file>");
    println!("Utility to copy files using splice(2)");
    println!("  -h        Display help.");
    println!("  -p, -a    Preserve file permissions.");
    println!("  -v        Enable verbose mode.");
}
